MON            (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)
DAY            ([ 0-3][0-9])
TIME           ([ 0-2][0-9]\:[0-6][0-9]\:[0-6][0-9])
TS             {MON}" "{DAY}" "{TIME}
O              ([0-9]|[0-9][0-9]|[0-1][0-9][0-9]|2[0-4][0-9]|25[0-5])
IP             {O}((\.{O}){3})
FIRST          [A-Za-z0-9]
MIDDLE         [-.A-Za-z0-9]
LAST           [A-Za-z0-9]
HN             ({FIRST}{MIDDLE}*{LAST}|{FIRST}{MIDDLE}*|{MIDDLE}*{LAST})

    #include <sys/types.h>
    #include <sys/socket.h>

    #include <netinet/in.h>

    #include <arpa/inet.h>
    #include <arpa/nameser.h>

    #include <ctype.h>
    #include <netdb.h>
    #include <stdio.h>
    #include <string.h>
    #include <time.h>
    #include <unistd.h>

    #ifdef notdef
    #ifdef HAVE_OS_PROTO_H
    #include "os-proto.h"
    #endif
    #endif

    #undef yywrap
    #ifdef FLEX_SCANNER
    #define YY_NO_UNPUT
    #endif

    int yywrap(void);
    int yylex(void);
    void addr(const char *, int);
    void name(const char *, int);
    void chain2(const char *, int, int);
    void chain3(const char *, int, int);
    void chainhack(const char *, int);
    void summary(void);

    int debug = 0;
    int timestamp = 0;
    int stats = 0;
    time_t lastts = 0;
    u_int lineno = 1;
    const char *file = "?";

%x ident eat junk

%%

<INITIAL>^{TS}" "	BEGIN(ident);

<INITIAL>^...		{
			BEGIN(junk);
			if (debug)
				fprintf(stderr, "%s:%d JUNK: \"%s",
				    file, lineno, yytext);
			}

<junk>.*		{
			if (debug)
				fprintf(stderr, "%s\"\n", yytext);
			}

<ident>{IP}" "		{
			addr(yytext, yyleng - 1);
			BEGIN(eat);
			}

<ident>{HN}" "		{
			name(yytext, yyleng - 1);
			BEGIN(eat);
			}

<ident>{IP}"/"{IP}"/"{IP}	{
			chain3(yytext, yyleng, 1);
			BEGIN(eat);
			}

<ident>{HN}"/"{IP}"/"{IP}	{
			chain3(yytext, yyleng, 0);
			BEGIN(eat);
			}

<ident>{IP}"/"{IP}	{
			chain2(yytext, yyleng, 1);
			BEGIN(eat);
			}

<ident>{HN}"/"{IP}	{
			chain2(yytext, yyleng, 0);
			BEGIN(eat);
			}

<ident>[^ ]+"/"{IP}	{
			/* Yech */
			chainhack(yytext, yyleng);
			BEGIN(eat);
			}

<eat>.*			/* eat */

<*>.			{
			/* XXX shouldn't see many of these */
			if (debug)
				fprintf(stderr, "{%s:%d %s}",
				    file, lineno, yytext);
			}

<*>\n			{
			++lineno;
			BEGIN(INITIAL);
			}

%%

/*
 * Copyright (c) 2007
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2007\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: logcount.l,v 1.18 2007/10/10 23:00:44 leres Exp $ (LBL)";
#endif

char *prog;

int targc;
char **targv;

extern char *optarg;
extern int optind, opterr;

#define HSIZE 8192			/* must be a power of two */

/* Address hash table */
struct htable {
	struct htable *next;
	char *name;
	int count;
	int state;
	u_int32_t addr;
} htable[HSIZE];

/* Simple name hash table (points back to address hash table) */
struct htable2 {
	struct htable2 *next;
	char *host;
	struct htable *htablep;
	int state;
} htable2[HSIZE];

#define STATE_FREE	0	/* not in use */
#define STATE_USED	1

/* Forwards */
int main(int, char **);
struct htable *cacheaddr(u_int32_t, int, const char *);
struct htable2 *cachehost(const char *, int);
inline struct htable2 *hash2(const char *);
inline struct htable *hash(u_int32_t);
struct htable *increment(u_int32_t, const char *);
struct htable *lookupaddr(u_int32_t);
struct htable2 *lookuphost(const char *);
void lowcase(char *);
volatile void usage(void);

/* Globals */
size_t numlst = 0;
size_t sizelst = 0;
struct in_addr *srclst = NULL;
struct in_addr *dstlst = NULL;
int addrcnt = 0;			/* number of gethostbyaddr() calls */
int hostcnt = 0;			/* number of gethostbyname() calls */
int totalhosts = 0;
int chainhackcnt = 0;

char azero[NS_IN6ADDRSZ];

int
main(argc, argv)
	int argc;
	char **argv;
{
	char *cp;
	int op;

	if (argv[0] == NULL)
		prog = "hf";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "dts")) != EOF)
		switch (op) {

		case 'd':
			++debug;
			break;

		case 't':
			++timestamp;
			break;

		case 's':
			++stats;
			break;

		default:
			usage();
		}

	/* Let yywrap() figure out if there are any arguments to open */
	targc = argc - optind;
	targv = &argv[optind];
	yyin = NULL;
	(void)yywrap();

	/* Process file opened by yywrap() or stdin if no arguments */
	if (yyin)
		yylex();

	summary();
	exit(0);
}

int
yywrap()
{
	static int didany = 0;

	/* Close file, if necessary */
	if (yyin) {
		if (yyin != stdin)
			(void)fclose(yyin);
		yyin = NULL;
	}

	/* Spin through arguments until we run out or successfully open one */
	while (targc > 0) {
		file = targv[0];
		--targc;
		++targv;
		++didany;
		if ((yyin = fopen(file, "r")) != NULL)
			return (0);
		perror(file);
	}
	if (!didany) {
		++didany;
		yyin = stdin;
		file = "<stdin>";
		return (0);
	}
	return (1);
}

void
addr(const char *txt, int len)
{
	struct in_addr a;
	char taddr[256];

	(void)snprintf(taddr, sizeof(taddr), "%.*s", len, txt);
	if (inet_aton(taddr, &a) == 0) {
		if (debug)
			fprintf(stderr, "%s:%d: inet_aton(%s) failed\n",
			    file, lineno, taddr);
		return;
	}
	(void)increment(a.s_addr, NULL);
}

/* Important that resolv.conf has a "search" config for all possible domains */
void
name(const char *host, int len)
{
	struct hostent *hp;
	char thost[256];
	char *fqdn;
	struct htable *p;
	struct htable2 *p2;

	static char lasthost[256] = "?";
	static char lastfqdn[256];
	static struct in_addr lastaddr;
	u_int32_t a;

	(void)snprintf(thost, sizeof(thost), "%.*s", len, host);

	/* Try simple name hash first */
	p2 = lookuphost(thost);
	if (p2 != NULL) {
		++p2->htablep->count;
		return;
	}

	hp = gethostbyname(thost);
	++hostcnt;
	if (hp == NULL) {
		if (debug > 1)
			fprintf(stderr, "%s:%d: gethostbyname(%s) failed\n",
			    file, lineno, thost);
		return;
	}
	lowcase(hp->h_name);
	fqdn = hp->h_name;

	if (hp->h_length != NS_INADDRSZ) {
		fprintf(stderr, "%s:%d: hp->h_length != NS_INADDRSZ (%d)\n",
		    file, lineno, hp->h_length);
		return;
	}

	if (hp->h_addrtype != AF_INET) {
		fprintf(stderr, "%s:%d: hp->h_addrtype != AF_INET (%d)\n",
		    file, lineno, hp->h_addrtype);
		return;
	}

	memmove(&a, hp->h_addr_list[0], NS_INADDRSZ);
	p = increment(a, fqdn);

	p2 = cachehost(thost, STATE_USED);
	p2->htablep = p;

	strcpy(lastfqdn, fqdn);
	strcpy(lasthost, thost);
	memmove(&lastaddr, hp->h_addr_list[0], sizeof(lastaddr));
}

/* Handles HN/IP and IP/IP */
void
chain2(const char *str, int len, int isip)
{
	const char *cp;
	struct in_addr a;

	cp = strchr(str, '/');
	if (cp == NULL) {
		if (debug)
			fprintf(stderr, "%s:%d: missing '/': %s\n",
			    file, lineno, str);
		return;
	}
	++cp;

	if (inet_aton(cp, &a) == 0) {
		/* fall back on hostname part of chain */
		if (debug)
			fprintf(stderr, "%s:%d: inet_aton(%s) failed\n",
			    file, lineno, cp);
		name(str, cp - str - 1);
		return;
	}

	if (ntohl(a.s_addr) == INADDR_LOOPBACK) {
		/* fall back on hostname part of chain */
		if (debug > 2)
			fprintf(stderr, "%s:%d: localhost: %.*s\n",
			    file, lineno, cp - str - 1, str);
		if (!isip)
			name(str, cp - str - 1);
		else
			addr(str, cp - str - 1);
		return;
	}
	(void)increment(a.s_addr, NULL);
}

/* Handles HN/IP/IP and IP/IP/IP (strip off the last IP and invoke chain2) */
void
chain3(const char *str, int len, int isip)
{
	const char *cp;
	char buf[512];

	cp = strrchr(str, '/');
	if (cp == NULL) {
		if (debug)
			fprintf(stderr, "%s:%d: missing '/': %s\n",
			    file, lineno, str);
		return;
	}
	len = cp - str;
	(void)snprintf(buf, sizeof(buf), "%.*s", len, str);
	chain2(buf, len, isip);
}

/*
 * Deal with cruft like:
 *
 *  "Oct  5 01:55:02 /Applications/Retrospect/131.243.10.24 6.1/Retro ..."
 *  "Oct  5 00:00:02 \tRev./128.3.28.153 2 LSI, Inc. 1064 found."
 */
void
chainhack(const char *str, int len)
{
	const char *cp;
	char temp[256];

	if (debug > 2)
		fprintf(stderr, "%s:%d: chainhack \"%.*s\"\n",
		    file, lineno, len, str);
	++chainhackcnt;
	cp = strrchr(str, '/');
	(void)snprintf(temp, sizeof(temp), "?%s", cp);
	chain2(temp, strlen(temp), 0);
}

/* Address required, fqdn is optional */
struct htable *
increment(u_int32_t a, const char *fqdn)
{
	const char *cp;
	struct htable *p;
	struct hostent *hp;
	struct in_addr ip;

	p = lookupaddr(a);
	if (p == NULL) {
		if (fqdn == NULL) {
			hp = gethostbyaddr(&a, NS_INADDRSZ, AF_INET);
			++addrcnt;
			if (hp == NULL) {
				ip.s_addr = a;
				cp = inet_ntoa(ip);
				if (debug > 1)
					fprintf(stderr,
					    "%s:%d: Can't translate: %s\n",
					    file, lineno, cp);
				fqdn = cp;
			} else {
				lowcase(hp->h_name);
				fqdn = hp->h_name;
				if (hp->h_length != NS_INADDRSZ) {
					fprintf(stderr, "%s:%d: hp->h_length"
					    " != NS_INADDRSZ (%d)\n",
					    file, lineno, hp->h_length);
					return (NULL);
				}

				if (hp->h_addrtype != AF_INET) {
					fprintf(stderr, "%s:%d: hp->h_addrtype"
					    " != AF_INET (%d)\n",
					    file, lineno, hp->h_addrtype);
					return (NULL);
				}
			}
		}
		p = cacheaddr(a, STATE_USED, fqdn);
		if (p == NULL) {
			fprintf(stderr, "increment: cacheaddr failed\n");
			abort();
		}
	}
	++p->count;
	return (p);
}

void
summary()
{
	int i, j;
	struct htable *p, *p2;
	char buf[132];
	struct tm *tp;
	time_t now;
	FILE *f;
	int maxcol;
	int emptybkts;

	f = stdout;
	if (timestamp) {
		(void)time(&now);
		tp = localtime(&now);
		strftime(buf, sizeof(buf), "%F %R", tp);
		fprintf(f, "# Updated: %s\n", buf);
	}

	if (stats) {
		fprintf(f, "# Total hosts: %d\n", totalhosts);
		if (stats > 1) {
			fprintf(f, "# gethostbyaddr() calls: %d\n", addrcnt);
			fprintf(f, "# gethostbyname() calls: %d\n", hostcnt);
			fprintf(f, "# chainhack() calls: %d\n", chainhackcnt);
			maxcol = 0;
			emptybkts = 0;
			for (p = htable, i = 0; i < HSIZE; ++p, ++i) {
				if (p->state == STATE_FREE) {
					++emptybkts;
					continue;
				}
				j = 0;
				for (p2 = p; p2 != NULL; p2 = p2->next)
					++j;
				if (maxcol < j)
					maxcol = j;
			}
			fprintf(f, "# Empty buckets: %d/%d\n",
			    emptybkts, HSIZE);
			fprintf(f, "# max collisions: %d\n", maxcol);
		}
	}

	for (p = htable, i = 0; i < HSIZE; ++p, ++i) {
		if (p->name == NULL)
			continue;
		for (p2 = p; p2 != NULL; p2 = p2->next) {
			if (inet_ntop(AF_INET, &p2->addr,
			    buf, sizeof(buf)) == NULL)
				strcpy(buf, "?");
			fprintf(f, "%s\t%s\t%d\n", p2->name, buf, p2->count);
		}
	}

}

/* Look hash table entry for address */
struct htable *
lookupaddr(u_int32_t a)
{
	struct htable *p;

	for (p = hash(a); p != NULL; p = p->next)
		if (a == p->addr)
			return (p);
	return (NULL);
}

/* Look hash table entry for a simple name */
struct htable2 *
lookuphost(const char *host)
{
	struct htable2 *p;

	for (p = hash2(host); p != NULL; p = p->next)
		if (p->host != NULL && strcmp(host, p->host) == 0)
			return (p);
	return (NULL);
}


inline struct htable *
hash(const u_int32_t a)
{
	u_int32_t h;

	h = (a & 0xff) ^ ((a >> 16) & 0xff);
	return (&htable[h & (HSIZE - 1)]);
}

struct htable2 *
hash2(const char *name)
{
	size_t len;
	u_int32_t h;

	len = strlen(name);
	if (len > sizeof(h))
		len = sizeof(h);
	h = 0;
	memmove(&h, name, len);
	return (&htable2[h & (HSIZE - 1)]);
}

struct htable *
cacheaddr(u_int32_t a, int state, const char *fqdn)
{
	struct htable *p, *p2;

	/* Don't cache zero */
	if (a == 0)
		return (NULL);

	/* Look for existing slot in hash table */
	for (p = hash(a); p != NULL; p = p->next)
		if (p->state != STATE_FREE && a == p->addr)
			break;

	/* Allocate a new slot */
	if (p == NULL) {
		p = hash(a);
		if (p->state != STATE_FREE) {
			/* Handle the collision */
			p2 = (struct htable *)malloc(sizeof(*p2));
			/* Lose, lose */
			if (p2 == NULL)
				return (NULL);
			memset((char *)p2, 0, sizeof(*p2));
			p2->next = p->next;
			p->next = p2;
			p = p2;
		}
	}

	/* Install new host */
	++totalhosts;
	p->addr = a;
	if (fqdn != NULL)
		p->name = strdup(fqdn);
	if (state != 0)
		p->state = state;
	if (p->state == STATE_FREE)
		abort();

	/* Return answer entry */
	return (p);
}

struct htable2 *
cachehost(const char *host, int state)
{
	struct htable2 *p, *p2;

	/* Look for existing slot in hash table */
	for (p = hash2(host); p != NULL; p = p->next)
		if (p->state != STATE_FREE && strcmp(host, p->host) == 0)
			break;

	/* Allocate a new slot */
	if (p == NULL) {
		p = hash2(host);
		if (p->state != STATE_FREE) {
			/* Handle the collision */
			p2 = (struct htable2 *)malloc(sizeof(*p2));
			/* Lose, lose */
			if (p2 == NULL)
				return (NULL);
			memset((char *)p2, 0, sizeof(*p2));
			p2->next = p->next;
			p->next = p2;
			p = p2;
		}
	}

	/* Install new host */
	p->host = strdup(host);
	if (state != 0)
		p->state = state;
	if (p->state == STATE_FREE)
		abort();

	/* Return answer entry */
	return (p);
}
void
lowcase(char *str)
{
	char *cp;

	for (cp = str; *cp != '\0'; ++cp)
		if (isupper(*cp))
			*cp = tolower(*cp);
}

volatile void
usage()
{
	extern char version[];

	(void)fprintf(stderr, "Version %s\n", version);
	(void)fprintf(stderr, "usage: %s [-dts] [file ...]\n", prog);
	exit(1);
}
